{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE BangPatterns #-}

{-
author: kstahl@mailbox.tu-berlin.de

compile with: ghc -O2 -fexcess-precision -funfolding-use-threshold=16 -o pclustering pclustering.hs
-}

module Main where

import Control.Monad
import System.Console.GetOpt
import System.IO
import System.FilePath
import System.Exit 
import System.Directory (createDirectoryIfMissing,doesFileExist)
import System.Environment (getArgs)
import Data.List
import Data.Time
import Data.Maybe (fromMaybe,isJust)
import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import qualified Data.Text.Read as TR
import qualified Data.Map as M
import Text.ParserCombinators.Parsec
import qualified Data.Foldable as Fold
import qualified Control.Applicative as CA
import qualified Data.IntMap as IM

data PatternSymbol = Box | Star deriving (Eq, Ord)

instance Show PatternSymbol where
  show Box = "_"
  show Star = "*"

data Alphabet = Discrete {-# UNPACK #-} !Int | Interval {-# UNPACK #-} !Double deriving (Read, Eq, Ord)
instance Show Alphabet where
  show (Interval a) = show a
  show (Discrete a) = show a

type Record = [Alphabet]
data MatchedPosition = StarField | DropField | DiscreteField {-# UNPACK #-} !Int | IntervalField {-# UNPACK #-} !Double deriving (Eq,Ord)

instance Show MatchedPosition where
  show StarField = show Star
  show (DiscreteField a) = show a
  show (IntervalField a) = show a
  show DropField = show Star

type Pattern = [PatternSymbol]
type MatchedRecord = [MatchedPosition]

data Evaluation = Evaluation {
  k :: !Int,
  usefulness :: !Double,
  weightedUsefulness :: !Double,
  noOfEquivalenceClasses :: !Double,
  avgEquivalenceClassSize :: !Double,
  noOfSuppressions :: !Int,
  weightedNoOfSuppressions :: !Int,
  sumOfSquaredECSize :: !Int,
  maximumECSize :: !Double,
  noOfColumns :: !Int,
  distcount :: !Double,
  size :: !Int,
  time :: String
}

instance Show Evaluation where
  show (Evaluation k use wuse nac aac ns wns ss mac nc dc size time) = "k: " ++ show k ++ "\nnumber of suppressions: " ++ show ns
    ++ "\nweighted number of suppressions: " ++ show wns
    ++ "\nnumber of equivalence classes (ec): " ++ show nac ++ "\nuseness: " ++ show use ++ "\nweighted useness: " ++ show wuse
    ++ "\navg. size of ec: " ++ show aac  ++ "\nsum of squared ec sizes: " ++ show ss ++ "\nmaximum ec size: " ++ show mac
    ++ "\nnumber of columns: " ++ show nc ++ "\nmax number of distinct elements: " ++ show dc ++ "\nsize of dataset: " ++ show size
    ++ "\n\n#" ++ time ++ " " ++ show k ++ " " ++ show ns ++ " " ++ show wns ++ " " ++ show nac ++ " " ++ show use ++ " " ++ show wuse ++ " " ++ show aac
    ++ " " ++ show ss  ++ " " ++ show mac ++ " " ++ show nc ++ " " ++ show dc ++ " " ++ show size ++ "\n"

cmpRecordsByPattern :: Pattern -> Record -> Record -> Bool
cmpRecordsByPattern [] [] [] = True
cmpRecordsByPattern (p:ps) (r1:r1s) (r2:r2s) = 
        case p of 
          Star -> cmpRecordsByPattern ps r1s r2s
          Box -> (r1 == r2) && cmpRecordsByPattern ps r1s r2s
cmpRecordsByPattern _ _ _ = False

orderRecordsByPattern :: Pattern -> Record -> Record -> Ordering
orderRecordsByPattern (p:ps) (r1:r1s) (r2:r2s)
      | p == Star = orderRecordsByPattern ps r1s r2s
      | r1 < r2 = LT
      | r1 > r2 = GT
      | otherwise = orderRecordsByPattern ps r1s r2s
orderRecordsByPattern _ _ _ = EQ

applyPattern :: Pattern -> Record -> MatchedRecord
applyPattern = zipWith (\x y -> if x == Box
                                then case y of
                                      Discrete a -> DiscreteField a
                                      Interval a -> IntervalField a
                                else StarField)

dropFields :: Record -> MatchedRecord
dropFields rec = replicate (length rec) DropField

extractComponents :: Pattern -> [(MatchedRecord,[Record])] ->  ([[Record]], [[Record]]) -> ([(MatchedRecord,[Record])], [Record])
extractComponents p imres (xs, ys) = (foldl' (\(!racc) x -> (applyPattern p (head x),x) : racc) imres xs, concat ys)                                  

-- | simple greedy heuristic: the records are sorted and grouped according to a given pattern - groups bigger or equal k build an equivalence class,
-- the rest is fed back into the function and tried with a different pattern
greedyHeuristic :: Int -> [Pattern] -> [Record] -> [(MatchedRecord,[Record])] -> [(MatchedRecord,[Record])]
greedyHeuristic _ [] _ res = res
greedyHeuristic _ _ [] res = res
greedyHeuristic !k (p:ps) list res
    | length list < k = (dropFields (head list),list):res
    | otherwise = 
        let (res2, rest) = extractComponents p res . partition (\x -> length x >= k) . groupBy (cmpRecordsByPattern p) $ sortBy (orderRecordsByPattern p) list
        in greedyHeuristic k ps rest res2

-- | creates all possible pattern permutations of a given length (n), the result is presorted from low cost (few * as possible) to high
createPatternPermutations :: Int -> [Pattern]
createPatternPermutations !n 
    | n <= 0 = []
    | otherwise = replicateM n [Box,Star]

createPatternPermutations2 :: Int -> [Pattern]
createPatternPermutations2 !c = concatMap helper $ zip (reverse [0..c]) [0..c]

helper :: (Int, Int) -> [Pattern]
helper (c,            0)          = [replicate c Box]
helper (0,            c)          = [replicate c Star]
helper (cUnderscores, cAsterisks) = map (Box :) (helper (cUnderscores - 1, cAsterisks))
                                 ++ map (Star :) (helper (cUnderscores, cAsterisks - 1))

-- | calculates the cost of a given Pattern Vector: number of stars (*) * weights (default: 1)
costOfPattern :: [Int] -> Pattern -> Int
costOfPattern [] pattern = foldl'(\(!acc) p -> case p of
                                                Star -> acc + 1
                                                _ -> acc
                                 ) 0 pattern
costOfPattern weights pattern = sum $ zipWith (\w p -> case p of
                                                        Star -> w
                                                        _ -> 0
                                              ) weights pattern


cmpPattern :: [Int] -> Pattern -> Pattern -> Ordering
cmpPattern weights p1 p2 = costOfPattern weights p1 `compare` costOfPattern weights p2

costOfMatchedRecord :: [Int] -> MatchedRecord -> Int
costOfMatchedRecord [] pattern = foldl'(\(!acc) p -> if p == StarField || p == DropField
                                                     then acc + 1
                                                     else acc
                                       ) 0 pattern
costOfMatchedRecord weights pattern = sum $! zipWith (\w p -> case p of
                                                              StarField -> w
                                                              DropField -> w
                                                              _ -> 0
                                                     ) weights pattern

evaluateResult :: Int -> [Int] -> Int -> [(MatchedRecord,[Record])] -> ([ColumnType],[ColumnData]) -> Mapping -> Int -> Double -> FilePath -> IO (String -> Evaluation)
evaluateResult !k weights !size res (b,coldata) mapping !nc !dcount fp = do
                                 let (noOfEC,(use,wuse)) = useness weights (b,coldata) res
                                 ((squared,suppr), (wsuppr, maxECS)) <- costs weights coldata res mapping fp
                                 return (Evaluation k use wuse noOfEC (fromIntegral size/noOfEC) suppr wsuppr squared maxECS nc dcount size)

-- | computes the squared size of the equivalence classes, the number of (weighted-)suppressions and the maximum equivalence class size
-- as a side effect the generalized result is written to a file
costs :: [Int] -> [ColumnData] -> [(MatchedRecord,[Record])] -> Mapping -> FilePath -> IO ((Int,Int),(Int,Double))
costs weights coldata res mapping fp = do
                  handle <- openFile fp WriteMode
                  res' <- foldM (\((!acc1,!acc2),(!acc3, !acc4)) (rs,rst) -> do
                                  let lrst = length rst
                                      !dlrst = fromIntegral lrst
                                  TIO.hPutStrLn handle $ T.intercalate "," (T.pack (show lrst) : generalize coldata rs mapping)
                                  return ((acc1 + (lrst ^ 2), acc2 + costOfMatchedRecord [] rs * lrst), (acc3 + costOfMatchedRecord weights rs * lrst, max acc4 dlrst))
                                ) ((0,0),(0,-1/0)) res
                  hClose handle
                  return res'

type Mapping = IM.IntMap T.Text

-- | function to parse the input file: discrete values are mapped to Integers, interval values are parsed as Doubles
parseFile :: String -> [ColumnType] -> Int -> IO (Maybe (Mapping, ([Record],Int)))
parseFile file b count = T.lines CA.<$> TIO.readFile file >>= \inp -> return (parseFile' inp b count)

parseFile' :: [T.Text] -> [ColumnType] -> Int -> Maybe (Mapping, ([Record],Int))
parseFile' [] _ _ = Nothing
parseFile' t@(t':_) b count' = foldM(\(!mapping,(rs,!count)) t -> do 
                                                                  let ts = T.split (== ',') t
                                                                      !tslen = length ts
                                                                  (mapping',r) <- parseEntry mapping (zip b ts)
                                                                  if count' /= tslen
                                                                  then Nothing
                                                                  else Just (mapping', (r:rs,count+1))
                                    ) ((M.empty,IM.empty),([],0)) t 
                              >>= \((_,m),(recs,count)) -> return (m,(recs,count))

parseField :: (M.Map T.Text Int, Mapping) -> (ColumnType,T.Text) -> Maybe ((M.Map T.Text Int, Mapping), Alphabet)
parseField (m,bm) (DiscreteT,s) = case lsm of
                                 Nothing -> Just ((M.insert s nkey m, IM.insert nkey s bm), Discrete nkey)
                                 Just !k -> Just ((m,bm), Discrete k)
                                where !lsm = M.lookup s m
                                      !nkey = M.size m
parseField (m,bm) (IntervalT,s) = case TR.double s of
                                    Right (!n,"")  -> Just ((m,bm), Interval n)
                                    _   -> Nothing

parseEntry :: (M.Map T.Text Int, Mapping) -> [(ColumnType,T.Text)] -> Maybe ((M.Map T.Text Int, Mapping), Record)
parseEntry m = Fold.foldrM(\inp (!mapping,rs) -> parseField mapping inp >>= \(mapping',r) -> return (mapping',r:rs)                                            
                          ) (m,[])

type ColumnData = Either (Double,Double) Double

-- | obtain column information, Left contains (min,max) of the interval values, Right: number of distinct discrete values in that particular column
extractMinMax :: [ColumnType] -> [Record] -> [ColumnData]
extractMinMax b res = foldl' f [] $ zip b (transpose res)
      where f acc (DiscreteT,r) = let !distcount = (fromIntegral . length . nub) r
                                  in Right distcount : acc
            f acc (IntervalT,r) = Left (minmax r) : acc

extractMinMaxInitial :: [ColumnType] -> [Record] -> (Double,[ColumnData])
extractMinMaxInitial b res = foldl' (\acc (t,r) -> let rnub = nub r
                                                       !dcount = fromIntegral $ length rnub
                                                   in f acc t dcount rnub
                                    ) (0,[]) $ zip b (transpose res)
      where f (!d,acc) DiscreteT !dcount _ = (max d dcount, Right dcount : acc)
            f (!d,acc) IntervalT !dcount r = (max d dcount, Left (minmax r) : acc)

minmax :: Record -> (Double,Double)
minmax = foldl'(\(!minacc,!maxacc) (Interval !i) -> (min minacc i, max maxacc i)) (1/0,-1/0)

-- | calculates the useness (see paper: Capturing Data Usefulness and Privacy Protection in K-Anonymisation; and function da)
useness :: [Int] -> ([ColumnType],[ColumnData]) -> [(MatchedRecord,[Record])] -> (Double,(Double,Double))
useness weights (b,coldata) m = let !avgweight = fromIntegral (sum weights) / fromIntegral (length weights)
                                    (!mlen,(!useness,!weightedUseness)) = 
                                            foldl'(\(!len,(!u,!wu)) (_,rs) -> 
                                                    let ecd = extractMinMax b rs
                                                        !u' = u + sum (zipWith (da 1 1) coldata ecd)
                                                        !wu' = if null weights
                                                               then u'
                                                               else wu + sum (zipWith3 (da avgweight) weights coldata ecd)      
                                                    in (len+1,(u',wu'))
                                                  ) (0,(0,0)) m
                                in (mlen,(useness / mlen, weightedUseness / mlen))

da :: Double -> Int -> ColumnData -> ColumnData -> Double
da !avgweight !weight (Left (!minDa,!maxDa)) (Left (!minVa,!maxVa)) = if maxDa - minDa /= 0
                                                                      then (maxVa - minVa) * (fromIntegral weight / avgweight) / (maxDa - minDa)
                                                                      else 0
da !avgweight !weight (Right !allD) (Right !allG) = allG * (fromIntegral weight / avgweight) / allD

patternFile = endBy line newline
line = sepBy symbol (char ',')

symbol = try parseBox <|> try parseStar

parseBox = do 
              char '_' <|> char '1'
              return Box

parseStar = do
              char '*' <|> char '0'
              return Star

data ColumnType = DiscreteT | IntervalT

discreteFile = sepBy1 symbol' (char ',')

symbol' = try parseDiscrete <|> try parseInterval

parseDiscrete = do 
              char 'd'
              return DiscreteT

parseInterval = do
              char 'i'
              return IntervalT

data Options = Options
     { optHelp     :: Bool
     , optPattern :: Maybe String
     , optK      :: Maybe [Int]
     , optWeight :: Maybe [Int]
     , optColTypes :: Maybe String
     } deriving (Eq,Show)

defaultOptions :: Options
defaultOptions = Options
     { optHelp    = False
     , optPattern = Nothing
     , optK     = Just [2,3,5,10]
     , optWeight     = Just []
     , optColTypes = Nothing
     }

options :: [OptDescr (Options -> Options)]
options =
  [ 
    Option "h" ["help"]
      (NoArg (\ opts -> opts { optHelp = True })) "print this help"
  ,    Option [] ["colTypes"]
      (OptArg (\ f opts -> opts { optColTypes = f })
              "d")
       "path to column types file (file format: single line csv with d for discrete, i for interval \
        \e.g.: i,d,i,d,i)"
  , Option [] ["k"]
      (OptArg ((\ f opts -> opts { optK = f }) . parseIntList . fromMaybe "") 
                "k")
         "list of k's (e.g.: 1,2,3), default: 2,3,5,10"
  , Option [] ["w"]
      (OptArg ((\ f opts -> opts { optWeight = f }) . parseIntList . fromMaybe "") 
                "w")
         "list of weights (e.g.: 1,2,3), default: 1, provide as\nmany weights as columns exist; \
         \cost of a pattern is the sum of its weights\n(patterns are sorted from lowest to highest cost) \
         \ \nhas no effect when used with --p"
  , Option [] ["p"]
      (OptArg (\ f opts -> opts { optPattern = f }) 
                "p")
         "path to pattern file (file format: * or 0 to drop entry,\n_ or 1 to keep, (e.g.: 0,_,*,1))"
  ]

parseOpts :: [String] -> IO (Options, [String])
parseOpts argv =
  case getOpt Permute options argv of
    (args,[fs],[]) -> do
        exists <- doesFileExist fs
        unless exists $ hPutStrLn stderr (failure fs) >> exitFailure
        unless (isJust . optColTypes $ nargs args) $ hPutStrLn stderr "column types not specified, supply --d=file option" >> exitFailure
        if optHelp (nargs args) then
          ioError (userError (usageInfo header options)) >> exitSuccess
        else return (nargs args, [fs])
    (_,_,errs) -> ioError (userError (concat errs ++ usageInfo header options))
  where header = "Usage: pclustering [OPTION...] file"
        failure file = "File \"" ++ file ++ "\" doesn't exist"
        nargs = foldl (flip id) defaultOptions

parseIntList :: String -> Maybe [Int]
parseIntList s = case reads ("[" ++ s ++ "]") of
                  [(l,"")] -> Just l
                  _ -> Nothing

-- | constructs the output of a given equivalence class with interval columns generalized
generalize :: [ColumnData] -> MatchedRecord -> Mapping -> [T.Text]
generalize coldata mrecs mapping = zipWith generalize' coldata mrecs
    where generalize' :: ColumnData -> MatchedPosition -> T.Text
          generalize' (Left (!mmin,!mmax)) StarField = T.pack ("[" ++ show mmin ++ ":" ++ show mmax ++ "]")
          generalize' _ (DiscreteField d) = mapping IM.! d
          generalize' _ (IntervalField i) = T.pack (show i)
          generalize' _ _ = "*"

runEval :: [Int] -> [Pattern] -> [Int] -> [Record] -> Int -> Mapping -> FilePath -> [ColumnType] -> Int -> NominalDiffTime -> IO ()
runEval weights plist ks input !nc m f discrete linecount start = mapM_ (\(!k) -> do 
                                                                            start2 <- getCurrentTime
                                                                            let gh = greedyHeuristic k plist input []
                                                                                ff = "output" </> (base ++ "_k=" ++ show k ++ ".csv")
                                                                            eval <- evaluateResult k weights linecount gh (discrete,ex) m nc dcount ff
                                                                            stop <- getCurrentTime
                                                                            let !t = diffUTCTime stop start2 + start
                                                                            print (eval (show t))
                                                                        ) ks
                                              where (!dcount,ex) = extractMinMaxInitial discrete input
                                                    base = takeBaseName f

getK :: Maybe [Int] -> IO [Int]
getK = return . fromMaybe (fail "invalid k-list supplied")

getPattern :: Maybe String -> Int -> [Int] -> IO [Pattern]
getPattern Nothing count [] = return $ createPatternPermutations2 count
getPattern Nothing count weights = return . sortBy (cmpPattern weights) $ createPatternPermutations count
getPattern (Just fp) count w = parseFromFile patternFile fp >>= \pinput ->
                                 case pinput of 
                                    Left err -> fail (show err)
                                    Right p -> do 
                                                  let plength = (nub . map length) p
                                                  when (length plength /= 1 || head plength /= count) $ fail "invalid patterns: no. of columns don't match"
                                                  return (if null w then p else sortBy (cmpPattern w) p)

getDiscrete :: Maybe String -> Int -> IO [ColumnType]
getDiscrete (Just fp) count = parseFromFile discreteFile fp >>= \dinput ->
                                case dinput of
                                  Left err -> fail (show err)
                                  Right d -> do
                                                  when (length d /= count) $ fail "invalid column description: no. of columns don't match"
                                                  return d

getWeights :: Maybe [Int] -> Int -> IO [Int]
getWeights Nothing _ = fail "invalid weights supplied (invalid format)"
getWeights (Just w) count = do when (not (null w) && count /= length w) $ fail "invalid weights supplied (length mismatch)"
                               return w

getInput' :: FilePath -> [ColumnType] -> Int -> IO (Mapping,([Record],Int))
getInput' fp b count = parseFile fp b count >>= \inp ->
                case inp of
                  Nothing -> fail "invalid input: no. of columns don't match"
                  (Just r) -> return r

peek :: FilePath -> IO Int
peek fp = do
          handle <- openFile fp ReadMode
          first <- TIO.hGetLine handle
          hClose handle
          let !count = length $ T.split (== ',') first
          return count

main :: IO ()
main = do
    (args, [file]) <- getArgs >>= parseOpts
    start <- getCurrentTime
    count <- peek file
    ct <- getDiscrete (optColTypes args) count    
    (mapping,(records,linecount)) <- getInput' file ct count
    k <- getK (optK args)
    weights <- getWeights (optWeight args) count
    patterns <- getPattern (optPattern args) count weights
    createDirectoryIfMissing True "output"
    stop <- getCurrentTime
    let !t = diffUTCTime stop start
    runEval weights patterns k records count mapping file ct linecount t