{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE OverloadedStrings #-}

import qualified Text.Parsec.Token as P
import Text.Parsec.Language (haskellDef)
import Text.ParserCombinators.Parsec
import Control.Monad
import System.Environment (getArgs)
import Data.List
import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import qualified Data.Text.Read as TR
import Control.DeepSeq

type Input = [T.Text]

data Alphabet = Text T.Text | Number !Int | FPNumber !Double deriving (Read, Eq, Ord)
instance Show Alphabet where
  show (Text a) = show a
  show (Number a) = show a
  show (FPNumber a) = show a

type Record = [Alphabet]

parseFile :: String -> IO [Record]
parseFile = liftM (map (parseEntry . T.split (== ',')) . T.lines) . TIO.readFile

parseNames :: String -> IO Input
parseNames = liftM T.lines . TIO.readFile

sanityCheck :: [Record] -> [Int]
sanityCheck = nub . map length

isNumber :: T.Text -> Either String Alphabet
isNumber s = case TR.decimal s of
                    Right (n,"")  -> Right (Number n)
                    _   -> case TR.double s of
                              Right (n,"")     -> Right (FPNumber n)
                              _   -> Left "No number found"

parseField :: T.Text -> Alphabet
parseField str = case isNumber str of
                 Left _ -> Text str
                 Right n -> n

parseEntry :: [T.Text] -> Record
parseEntry = map parseField

parseInput :: [[T.Text]] -> [Record]
parseInput = map (map parseField)

isNumber' :: Alphabet -> Bool
isNumber' n = case n of
              Text _ -> False
              _ -> True

initialExtractMinMax :: [Record] -> Bool -> [Either (Double,Double) Double]
initialExtractMinMax [] _ = []
initialExtractMinMax res b = extractMM blist tres
                           where tres = transpose res
                                 blist = if b then
                                         replicate (length $ head res) False
                                         else map (all isNumber') tres

{- Left contains (min,max) and no of distinct values of the interval values, Right: no of the distinct discrete values-}
extractMM :: [Bool] -> [Record] -> [Either (Double,Double) Double]
extractMM [] _ = []
extractMM _ [] = []
extractMM (b:bs) (r:rs) = if b
                          then Left (minmax r (1/0,-1/0)) : extractMM bs rs
                          else Right distcount : extractMM bs rs
                          where !distcount = (fromIntegral . length . nub) r
  

minmax :: Record -> (Double,Double) -> (Double,Double)
minmax [] acc = acc
minmax (r:rs) (!minacc,!maxacc) = case r of
                        (Number a) -> minmax rs (min minacc (fromIntegral a), max maxacc (fromIntegral a))
                        (FPNumber a) -> minmax rs (min minacc a, max maxacc a)
                        _ -> minmax rs (minacc, maxacc)

da :: Either (Double,Double) Double -> InputData -> Double
da !dages dagroup = case dages of
        Left (!minDa,!maxDa) -> if maxDa - minDa /= 0
                                then discreteStepWidth dagroup / (maxDa - minDa)
                                else 0
        Right (!d) -> discreteStepWidth dagroup / d

da' :: Double -> Int -> Either (Double,Double) Double -> InputData -> Double
da' !avgweight !weight !dages !dagroup  = case dages of
        Left (!minDa,!maxDa) -> if maxDa - minDa /= 0
                                then discreteStepWidth dagroup * (fromIntegral weight / avgweight)/ (maxDa - minDa)
                                else 0
        Right (!d) -> discreteStepWidth dagroup * (fromIntegral weight / avgweight) / d

useness :: [Int] -> [Either (Double,Double) Double] -> [IRecord] -> (Double,Double)
useness weights ges ilst = (u/len, w/len)
                  where avgweight :: Double
                        !avgweight = fromIntegral (sum weights) / fromIntegral (length weights)
                        len = fromIntegral (length ilst)
                        (u,w) = foldl'(\((!acc,!wacc)) r -> let unweighted = sum (zipWith da ges r)
                                                                weighted = if null weights
                                                                           then unweighted
                                                                           else sum (zipWith3 (da' avgweight) weights ges r)
                                                            in (acc+unweighted, wacc+weighted)
                                      ) (0,0) ilst

data InputData = Interval {low :: {-# UNPACK #-} !Double, high ::  {-# UNPACK #-} !Double}
              | Discrete { n :: {-# UNPACK #-} !Int, ident :: String} deriving (Ord, Eq)

type IRecord = [InputData]
instance Show InputData where
    show !(Interval low high) = "[" ++ show low ++ ":" ++ show high ++ "]"
    show !(Discrete n ident) = show n ++ " " ++ ident

discreteStepWidth :: InputData -> Double
discreteStepWidth (Interval !low !high) = high - low
discreteStepWidth (Discrete !n _) = fromIntegral n

parseNumber = naturalOrFloat >>=
                    \num -> case num of 
                        Left (!i) -> return $! fromIntegral i
                        Right (!d) -> return d

lexer = P.makeTokenParser haskellDef 
naturalOrFloat = P.naturalOrFloat lexer
decimal = P.decimal lexer

csvFile = endBy line newline
line = sepBy cell (char ',')

cell = try closed <|> try open <|> try halfL <|> try halfR <|> try parseDiscrete

parseN = decimal >>= 
                  \(!num) -> return $! fromInteger num

parseDiscrete = do
                  char '#'
                  !val <- parseN
                  char '#'
                  !ident <- many (noneOf ",\n")
                  return $! Discrete val ident

closed = do 
            char '['
            !low <- parseNumber
            char ':'
            !high <- parseNumber
            char ']'
            when (low > high) $ fail("invalid interval: [" ++ show low ++ ":" ++ show high ++ "]")
            return $! Interval low high

open = do 
            char '('
            !low <- parseNumber
            char ':'
            !high <- parseNumber
            char ')'
            when (low == high || low > high) $ fail("invalid interval: (" ++ show low ++ ":" ++ show high ++ ")")
            return $! Interval (low + 1) (high - 1)

halfL = do 
            char '('
            !low <- parseNumber
            char ':'
            !high <- parseNumber
            char ']'
            when (low > high) $ fail("invalid interval: (" ++ show low ++ ":" ++ show high ++ "]")
            return $! Interval (low + 1) high

halfR = do 
            char '['
            !low <- parseNumber
            char ':'
            !high <- parseNumber
            char ')'
            when (low > high) $ fail("invalid interval: [" ++ show low ++ ":" ++ show high ++ ")")
            return $! Interval low (high - 1)

parseCSV = parse csvFile "Error"

costs :: [[IRecord]] -> ((Int,Int),(Double,Double))
costs = foldl' (\((!acc1,!acc2),(!acc3, !acc4)) rst ->
                                  let lrst = length rst
                                      dlrst = fromIntegral lrst 
                                  in ((acc1 + (lrst ^ 2), acc2 + lrst),(acc3 + 1, max acc4 dlrst))
                ) ((0,0),(0,-1/0))

process :: [Either (Double,Double) Double] -> [IRecord] -> String
process ex lst = show use ++ " " ++ show wuse ++ " " ++ show sizeACC ++ " " ++ show (fromIntegral size / sizeACC) ++ " " ++ show quadSum ++ " " ++ show maxAccSize
              where inp = (group . sort) lst
                    weights = []
                    (use,wuse) = useness weights ex $ map head inp
                    ((!quadSum, !size),(!sizeACC,!maxAccSize)) = costs inp

main = do
    (file:mondrian:discrete:_) <- getArgs
    recs <- parseFile file
    input <- parseFromFile csvFile mondrian
    let d = read discrete :: Bool
    let x = initialExtractMinMax recs d
    let !y = rnf x `deepseq` x
    case input of
        Left _ -> putStrLn "File corrupt"
        Right inp -> putStrLn $ process y inp 