package methods;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.LinkedList;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.lazy.IB1;
import weka.classifiers.lazy.IB1_Anon;
import weka.core.Instances;
import anonymizer.Anonymizer;
import anonymizer.Configuration;

/**
 * This class creates an experiment scenario for the IB1_Anon class.
 * The following parameters are necessary to create a scenario:
 * <ul>
 * <li>Anonymization parameters within a configuration file (see toolbox manual for details,
 * program arguments valid for the toolbox are valid here as well)
 * <li>Test scenario parameter: 00 implies training and testing over original data, 01 implies 
 * training over original, testing over anonymized data, 10 vice versa, and finally 11 implies
 * both training and testing over anonymized data (default value is set to 11).
 * <li>ARFF header file: The header should contain all necessary information upto and including
 * the &#64;DATA pointer. If anonymization parameters contain any identifier attributes, include 
 * those as well (identifiers will be removed automatically).
 * <li>Cross validation (optional): Number of folds for cross-validation (default value is set 
 * to 2). 
 * </ul>
 * Cross-validation is performed manually, since we need training and test datasets from
 * either anonymized or original data depending on the test scenario.
 */
public class ExpScenario {
	/**
	 * Runs an experiment scenario according to the specified arguments
	 * @param args Vector of arguments
	 * @throws Exception
	 */
	public static void runExp(String[] args) throws Exception{
		//test scenario params
		String testScenario = "11"; //00, 01, 10, 11
		String descriptor = null; //arff header file
		int crossVal = 2; //some integer value
		
		//update from arguments
		int index = -1;
		if( (index = getOption("-scenario", args)) >= 0) {
			if(args[index].compareTo("00") != 0 && args[index].compareTo("10") != 0 
				&& args[index].compareTo("01") != 0 && args[index].compareTo("11") != 0) {
				System.out.println("WARNING: Invalid test scenario, will assume 11");
			} else {
				testScenario = args[index];
			}
		}
		if( (index = getOption("-arff", args)) >= 0) { //not optional
			descriptor = args[index];
			File f = new File(descriptor);
			if(!f.exists()) {
				throw new Exception("ARFF header file " + descriptor + " not found!!!");
			}
		} else {
			throw new Exception("ARFF header file not specified (use option -arff)!!!");
		}
		if( (index = getOption("-crossval", args)) >= 0) {
			crossVal = Integer.parseInt(args[index]);
		}
		if(crossVal < 2) {
			System.out.println("WARNING: At least 2 folds required for " +
					"cross-validation. Setting the value to 2");
			crossVal = 2;
		}

		//anonymization params - make sure that output format is not anatomy
		Configuration conf = new Configuration(args);
		if(conf.outputFormat == Configuration.OUTPUT_FORMAT_ANATOMY) {
			String message = "Anatomy output format is not supported (do not use program " +
				"arguments to set outputFormat)!!!";
			throw new Exception(message);
		}
		if(conf.anonMethod == Configuration.METHOD_MONDRIAN) {
			String message = "Anonymization methods that generate their own VGHs (e.g., Mondrian)" +
				" cannot be used!!!";
			throw new Exception(message);
		}
		
		//identifier attributes will be handled cautiously
		int[] idFields = null;
		if(conf.idAttributeIndices.size() > 0) {
			idFields = new int[conf.idAttributeIndices.size()];
			for(int i = 0; i < idFields.length; i++) {
				idFields[i] = conf.idAttributeIndices.get(i);
			}
		}
		
		//fetch indices of quasi-identifier attributes
		int[] qidFields = new int[conf.qidAtts.length];
		for(int i = 0; i < qidFields.length; i++) {
			qidFields[i] = conf.qidAtts[i].index;
		}
		
		//remove tuples with missing values
		FileIO.RemoveRecWithMissingValues(conf.inputFilename, "temp1.tmp");
		//remove identifier attributes manually
		if(idFields != null) {
			//we should first re-adjust quasi-identifier indices
			for(int i = 0; i < conf.qidAtts.length; i++) {
				for(int j = 0; j < idFields.length; j++) {
					if(idFields[j] < conf.qidAtts[i].index) {
						conf.qidAtts[i].index--;
					}
				}
			}
			//then re-adjust sensitive indices
			for(int i = 0; i < conf.sensitiveAtts.length; i++) {
				for(int j = 0; j < idFields.length; j++) {
					if(idFields[j] < conf.sensitiveAtts[i].index) {
						conf.sensitiveAtts[i].index--;
					}
				}
			}
			//now remove the id attributes from the file
			FileIO.AttributeRemoval("temp1.tmp", "temp2.tmp", idFields);
			FileIO.RenameFile("temp2.tmp", "temp1.tmp");
			conf.idAttributeIndices = new LinkedList<Integer>();
		}
		conf.inputFilename = "temp1.tmp";
		
		//anonymize
		Anonymizer.anonymizeDataset(conf);
		
		//get header
		String header = FileIO.GetARFFHeader(descriptor, idFields, qidFields);

		//prepare cross-validation datasets
		for(int i = 1; i <= crossVal; i++){
			String fname1 = "orig"+i+".rawdata";
			String fname2 = "orig"+i+".rawtest";
			FileIO.PrepareCrossValidationData(conf.inputFilename, fname1, fname2, crossVal, i);
			FileIO.AddStringToFile(fname1, header);
			FileIO.AddStringToFile(fname2, header);
			fname1 = "anon"+i+".rawdata";
			fname2 = "anon"+i+".rawtest";
			FileIO.PrepareCrossValidationData(conf.outputFilename, fname1, fname2, crossVal, i);
			FileIO.AddStringToFile(fname1, header);
			FileIO.AddStringToFile(fname2, header);
		}
		
		//train and test crossVal times
		double[] stats = {0,0,0,0};
		for(int i = 0; i < crossVal; i++) {
			System.out.print(".");
			String trainData, testData;
			if(testScenario.compareTo("00")==0) {
				trainData = testData = "orig";
			}
			else if(testScenario.compareTo("11")==0) {
				trainData = testData = "anon";
			}
			else if(testScenario.compareTo("01")==0) {
				trainData = "orig";
				testData = "anon";
			}
			else {//if(testCase == "10")
				trainData = "anon";
				testData = "orig";
			}
			trainData = trainData + (i+1) + ".rawdata";
			testData = testData + (i+1) + ".rawtest";
			//Original data set in ARFF format in the file with filename anonInput
			Instances train = new Instances(new BufferedReader(new FileReader(
					trainData)));
			train.setClassIndex(train.numAttributes() - 1);
			//Anonymized data set in ARFF format in the file with filename anonOutput
			Instances test = new Instances(new BufferedReader(new FileReader(
					testData)));
			test.setClassIndex(test.numAttributes() - 1);
			
			Classifier nn;
			if(testScenario == "00") { //use original IB1
				nn = new IB1();
			} else {
				nn = new IB1_Anon(conf);
			}
			nn.buildClassifier(train);
			Evaluation eval = new Evaluation(train);
			eval.evaluateModel(nn, test);
			stats[0] += eval.correct();
			stats[1] += eval.incorrect();
			stats[2] += eval.unclassified();
			stats[3] += eval.numInstances();
		}
		//output results
		System.out.println();
		System.out.println("Correct: " + stats[0]);
		System.out.println("Incorrect: " + stats[1]);
		System.out.println("Unclassified: " + stats[2]);
		System.out.println("Total: " + stats[3]);
		System.out.println("Accuracy: " + (stats[0]/stats[3]));
		
		//delete temporary files create during the experiment
		deleteFile("temp1.tmp");
		for(int i = 1; i <= crossVal; i++){
			deleteFile("orig"+i+".rawdata");
			deleteFile("orig"+i+".rawtest");
			deleteFile("anon"+i+".rawdata");
			deleteFile("anon"+i+".rawtest");
		}
	}
	
	/**
	 * Deletes the file with the specified name
	 * @param fname path to the file
	 */
	private static void deleteFile(String fname) {
		File f = new File(fname);
		if(f.exists()) {
			f.delete();
		}
	}
	
	/**
	 * Gets the specified option
	 * @param option Option value to be searched
	 * @param args List of arguments
	 * @return Option as specified, or null if not set
	 */
	private static int getOption(String option, String[] args) {
		for(int i = 0; i < args.length; i++) {
			if(args[i].compareToIgnoreCase(option) == 0 && i+1 < args.length) {
				return i+1;
			}
		}
		return -1;
	}
	
	public static void main(String[] args) {
		//String argument = "-method datafly -config config.xml -scenario 01 -arff header.txt";
		//args = argument.split(" ");
		try {
			ExpScenario.runExp(args);
		} catch(Exception e) {e.printStackTrace();}
	}
}
